/*-
 * Copyright (C) @BABOLO  2002 Dec 04
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2006 Feb 21\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: cmd_trans.c,v 1.15 2007/12/02 01:21:15 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <poll.h>
#include <mife.h>
#include <err.h>
#include <multilar.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

int
pgoblin_wait(pgoblin_main *options, pgoblin_r *r) {
    int ex = EX_OK;
    pgoblin_job *j;
    size_t i;

    ifBLIN_QV4(options->flags) fprintf(stderr, "+pgoblin_wait %d\n", r[PGO_CJOB].r);
    if  (r[PGO_CJOB].g.j) {
        for (i = 0; i <  MULAR_NEXT(*r[PGO_CJOB].g.j); i++) {
            j = mular_getix(*r[PGO_CJOB].g.j, i);
            if  (!(j->pid)) {
                ifBLIN_QV1(options->flags) warnx("No pid in %d", j->cmdn);
                errno = EINVAL;
                ex = EX_SOFTWARE;
            } else if ((j->cmdn >= 0) && (j->status == PGOBLIN_NOSTATUS)) {
                if  ((ex = waitpid(j->pid, &(j->status), 0)) < 0) {
                    ifBLIN_QV1(options->flags) warn("Pid %d in %d", j->pid, j->cmdn);
                    ex = EX_OSERR;
                } else if (ex == j->pid) {
                    ex = EX_OK;
                } else {
                    ifBLIN_QV1(options->flags) warn("Pid %d<>%d in %d", ex, j->pid, j->cmdn);
                    ex = EX_OSERR;
    }   }   }   }
    ifBLIN_QV4(options->flags) fprintf(stderr, "-pgoblin_wait %d\n", ex);
    return(ex);
}

int
pgoblin_trap(pgoblin_main *options, pgoblin_r *r) {
    options->ctrap = options->ppoint + 1;
    options->exeflags |= PGOBLIN_TRAPDEF;
    options->lexlevel++;
    return(EX_OK);
}

int
pgoblin_call(pgoblin_main *options, pgoblin_r *r) {
    int ex = EX_UNAVAILABLE;
    return(ex);
}

int
pgoblin_close(pgoblin_main *options, pgoblin_r *r) {
    return(EX_OK);
}

int
pgoblin_begin(pgoblin_main *options, pgoblin_r *r) {
    PGresult *res;

    ifBLIN_QV4(options->flags) fprintf(stderr, "+pgoblin_begin\n");
    if  (!(R_CONN->intran)) {
        PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1, R_CONN->conn);
        PQclear(res);
        R_CONN->intran++;
    }
    ifBLIN_QV4(options->flags) fprintf( stderr, "-pgoblin_begin %d\n", 0);
    return(0);
}

int
pgoblin_end(pgoblin_main *options, pgoblin_r *r) {
    PGresult *res;

    ifBLIN_QV4(options->flags) fprintf(stderr, "+pgoblin_end\n");
    if  ((R_CONN->intran)) {
        PQEXEC(res, "END", PGRES_COMMAND_OK, -1, R_CONN->conn);
        PQclear(res);
        R_CONN->intran--;
    }
    ifBLIN_QV4(options->flags) fprintf( stderr, "-pgoblin_end %d\n", 0);
    return(0);
}

int
pgoblin_listen(pgoblin_main *options, pgoblin_r *r) {
    int       sock;
    struct pollfd pfd;
    int timeout, i, e, ex = EX_OK;
    int64_t tout;
    PGnotify *notify;
    struct timeval t;
    union {
        int64_t itime;
        u_int32_t d[2];
    } dbtime;
    typedef union {
        u_int64_t p;
        u_int32_t d[2];
    } pmatsemit;

    ifBLIN_QV4(options->flags) fprintf( stderr, "+pgoblin_listen %d %d %d =%s~\n"
                                      , r[PGO_CCTL].r, r[PGO_CCON].r, r[PGO_CIN].r
                                      , (char*)(IO_CTL->text)
                                      );
    if  ((sock = PQsocket(R_CONN->conn)) < 0) {
        ex = EX_UNAVAILABLE;
        errno = EINVAL;
    } else if (  !IO_IN->text
              || !(IO_IN->flags & PGOBLIN_BINPARM)
              || (IO_IN->length < sizeof(pmatsemit))
              ) {
        ex = EX_DATAERR;
        errno = EINVAL;
    } else {
        /* Should be 64 bit time wait to in usec from 1970-01-01 00:00:00 UTC *
         * (( date_part('epoch', t)                                           *
         *  + date_part('second', t)                                          *
         *  - floor(date_part('second', t))                                   *
         *  ) * 1000000                                                       *
         * )::int8                                                            *
         **********************************************************************/
        dbtime.d[1] = ntohl(((pmatsemit*)(IO_IN->text))->d[0]);
        dbtime.d[0] = ntohl(((pmatsemit*)(IO_IN->text))->d[1]);
        bzero(&pfd, sizeof(pfd));
        pfd.fd = sock;
        pfd.events = POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI;
        for (e = 1; e && !ex;) {
            if  (gettimeofday(&t, NULL)) {
                ex = EX_OSERR;
                break;
            }
            tout = dbtime.itime - ((int64_t)1000000LL * t.tv_sec + t.tv_usec);
            if  (tout <= 0) break;
            else if ((tout / 1000) > options->maxtmout) {
                timeout = options->maxtmout;
                dbtime.itime = ((int64_t)1000LL * options->maxtmout)
                           + ((int64_t)1000000LL * t.tv_sec)
                           + t.tv_usec
                           ;
            } else timeout = tout / 1000;
            ifBLIN_QV4(options->flags) fprintf( stderr, "LISTEN=%s~ fd %d for %d ms\n"
                                              , (char*)IO_CTL->text, sock, timeout
                                              );
            if  ((i = poll(&pfd, 1, timeout)) < 0) {
                ex = EX_OSERR;
                break;
            }
            PQconsumeInput(R_CONN->conn);
            while ( (notify = PQnotifies(R_CONN->conn))) {
                ifBLIN_QV4(options->flags)
                    fprintf(stderr, "Received NOTIFY=%s~\n", notify->relname);
                if  (!strcmp(notify->relname, IO_CTL->text)) e = 0;
                PQfreemem(notify);
    }   }   }
    ifBLIN_QV4(options->flags) fprintf( stderr, "-pgoblin_listen %d\n", ex);
    return(ex);
}
