/*-
 * Copyright (C) @BABOLO  2002 Dec 23
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 23\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: execute.c,v 1.31 2007/12/14 05:06:53 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/wait.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <multilar.h>
#include <libpq-fe.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

#define PRINTNOTNULL(A) (!(A) ? "" : (A))
#define PRINTANDNULL(A) PRINTNOTNULL(A), ((A) ? "~" : "")
#define PQSTATNDNULL(A) PRINTANDNULL(PQparameterStatus(options->conn[r].conn, (A)))
#define ALIGN8(A)  popt[i].A, (strlen(popt[i].A) < 8 ? "		" : "	")
#define ALIGN16(A) popt[i].A                                                                          \
      , ((strlen(popt[i].A) < 8) ? "		" : (strlen(popt[i].envvar) < 16) ? "	" : " ")

static void
NoticeProcessor(void *options, const char *message) {
    ifBLIN_QV1(((pgoblin_main *)options)->flags) fprintf(stderr, "%s", message);
}

int
pgoblin_shurecon(pgoblin_main *options, int r) {
    PQconninfoOption *popt;
    int i, ex = EX_OK;
    PGresult *res;

    if  (  options->conn[r].conn != NULL
        && PQstatus(options->conn[r].conn) == CONNECTION_OK
        ) return(0);
    ifBLIN_QV5(options->flags) {
        fprintf( stderr
               , "Connection defaults:\n"
                 "size=chr keyword	envvar		compiled val		label\n"
               );
        popt = PQconndefaults();
        for (i = 0; popt[i].keyword; i++) {
            fprintf( stderr
                   , "%4d=%s~	%s%s%s%s%s%s%s%s%s\n"
                   , popt[i].dispsize
                   , popt[i].dispchar
                   , ALIGN8(keyword)
                   , ALIGN16(envvar)
                   , PRINTNOTNULL(popt[i].compiled)
                   , !(popt[i].compiled) ? "	" : strlen(popt[i].compiled) < 7 ? "~	" : "~ "
                   , PRINTNOTNULL(popt[i].val)
                   , !(popt[i].val)
                   ? "		"
                   : strlen(popt[i].val) < 7 ? "~		" : "~	"
                   , popt[i].label
                   );
        }
        PQconninfoFree(popt);
            fprintf( stderr
                   , "PGREALM=%s%s PGDATESTYLE=%s%s PGTZ=%s%s PGCLIENTENCODING=%s%s PGGEQO=%s%s\n"
                   , PRINTANDNULL(getenv("PGREALM"))
                   , PRINTANDNULL(getenv("PGDATESTYLE"))
                   , PRINTANDNULL(getenv("PGTZ"))
                   , PRINTANDNULL(getenv("PGCLIENTENCODING"))
                   , PRINTANDNULL(getenv("PGGEQO"))
                   );
    }
    options->conn[r].conn = PQsetdbLogin( options->conn[r].host, options->conn[r].port, NULL, NULL
                                        , options->conn[r].dbname, options->conn[r].username, NULL
                                        );
    options->conn[r].flags |= PGOBLIN_CLOSECONN;
    ifBLIN_QV5(options->flags) {
        fprintf(stderr, "PQoptions=%s~\n", PQoptions(options->conn[r].conn));
        fprintf( stderr
               , "PQdb=%s~ PQuser=%s~ PQpass=%s%s "
                 "PQhost=%s%s PQport=%s%s "
                 "PQtty=%s%s\n"
               , PQdb(options->conn[r].conn), PQuser(options->conn[r].conn)
               , PRINTANDNULL(PQpass(options->conn[r].conn))
               , PRINTANDNULL(PQhost(options->conn[r].conn))
               , PRINTANDNULL(PQport(options->conn[r].conn))
               , PRINTANDNULL(PQtty(options->conn[r].conn))
               );
        fprintf( stderr
               , "server_version=%s%s "
                 "client_encoding=%s%s "
                 "is_superuser=%s%s "
                 "session_authorization=%s%s "
                 "DateStyle=%s%s\n"
               , PQSTATNDNULL("server_version")
               , PQSTATNDNULL("client_encoding")
               , PQSTATNDNULL("is_superuser")
               , PQSTATNDNULL("session_authorization")
               , PQSTATNDNULL("DateStyle")
               );
        fprintf( stderr
               , "PQprotocolVersion=%d PQsocket=%d PQbackendPID=%d\n"
               , PQprotocolVersion(options->conn[r].conn)
               , PQsocket(options->conn[r].conn)
               , PQbackendPID(options->conn[r].conn)
               );
    }
    ifBLIN_QV5(options->flags) fprintf(stderr, "PQstatus %d\n", PQstatus(options->conn[r].conn));
    PQsetNoticeProcessor(options->conn[r].conn, NoticeProcessor, options);
    if  (PQstatus(options->conn[r].conn) != CONNECTION_BAD) {
        if  (  options->conn[r].flags & PGOBLIN_NEEDINTIME
            && !(options->conn[r].flags & (PGOBLIN_INT64TIME | PGOBLIN_FLOATTIME))
            ) {
            if  (!options->conn[r].intran) {
/*          if  (PQtransactionStatus(options->conn[r].conn) != PQTRANS_INTRANS) { */
                PQEXEC(res, "BEGIN", PGRES_COMMAND_OK, -1, options->conn[r].conn);
                PQclear(res);
            }
            PQEXEC( res
                  , "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR "
                    "SELECT '2000-01-01 00:00:01'::timestamp;"
                    "FETCH 1 FROM pgoblin"
                  , PGRES_TUPLES_OK, -1
                  , options->conn[r].conn
                  );
            if  (  PQgetisnull(res, 0, 0)
                || (PQgetlength(res, 0, 0) < 8)
                ) {
                ex = EX_DATAERR;
                errno = EINVAL;
            } else {
                if  (*((u_char*)(PQgetvalue(res, 0, 0))) == 077) {
                    options->conn[r].flags |= PGOBLIN_FLOATTIME;
                    ifBLIN_QV3(options->flags)
                        fprintf(stderr, "conn %d float timestamp format\n", r);
                } else if (*((u_char*)(PQgetvalue(res, 0, 0))) == 0) {
                    options->conn[r].flags |= PGOBLIN_INT64TIME;
                    ifBLIN_QV3(options->flags)
                        fprintf(stderr, "conn %d int64 timestamp format\n", r);
                } else {
                    ifBLIN_QV1(options->flags)
                        fprintf(stderr, "conn %d unknown timestamp format\n", r);
                    ex = EX_DATAERR;
                    errno = EFTYPE;
            }   }
            PQclear(res);
            PQEXEC(res, "CLOSE pgoblin", PGRES_COMMAND_OK, -1, options->conn[r].conn);
            PQclear(res);
            if  (!options->conn[r].intran) {
                PQEXEC(res, "END", PGRES_COMMAND_OK, -1, options->conn[r].conn);
                PQclear(res);
        }   }
        return(0);
    }
    ifBLIN_QV1(options->flags)
        warnx("Connection to PostgreSQL failed: %s\n", PQerrorMessage(options->conn[r].conn));
    return(1);
}

static int
pgoblin_connown(pgoblin_main *options) {
    int ex = EX_OK; /* All errors must be < 0 */
    ssize_t g;

    ifBLIN_QV5(options->flags) fprintf(stderr, "+pgoblin_connown\n");
    for (g = 0; g < PGOBLIN_REGSIZE; g++) {
        options->conn[g].flags &= ~PGOBLIN_CLOSECONN;
    }
    ifBLIN_QV5(options->flags) fprintf(stderr, "-pgoblin_connown %d\n", ex);
    return(ex);
}

static size_t u12[] = {341, 512, 512};

pgoblin_main *
pgoblin_init(u_int32_t flags, int outfd, mife_descriptor *in, pgoblin_prog *pgm) {
    pgoblin_main *options = NULL;

    ifBLIN_QV3(flags) fprintf(stderr, "+pgoblin_init options=%s\n", options ? "OK" : "NULL");
    if  (pgm && strncmp(pgm->id, "#pGoblin" VERS, PGOBLIN_STRING_ID_LEN)) {
        errno = EINVAL;
        ifBLIN_QV1(flags) warnx("Illegal pgoblin program");
    } else if (!(options = malloc(sizeof(pgoblin_main)))) {
        errno = ENOMEM;
    } else {
        pgoblin_job *j;

        bzero(options, sizeof(pgoblin_main));
        options->flags = flags;
        strncpy(options->vers, VERS, 4);
        IO_RG0.flags |= PGOBLIN_OUTSET;
        IO_RG0.onu = outfd;
        IO_RG0.wri = mife_writ;
        IO_RG0.mife = in;
        options->maxtmout = 3600000;
        strncpy(options->delim, "#", 3);
        options->job[0] = mular_create(MULAR_ZERO, 3, 12, u12);
        j = mular_add(options->job[0]);
        j->cmdn = -1;
        j->pid = getppid();
        j->status = PGOBLIN_NOSTATUS;
        j = mular_add(options->job[0]);
        j->cmdn = -1;
        j->pid = getpid();
        j->status = PGOBLIN_NOSTATUS;
        if  (pgm) {
            options->pgm = pgm;
        } else if (( options->pgm
                   = malloc(sizeof(pgoblin_tuple) * PGOBLIN_PROGSIZE + sizeof(pgoblin_prog))
                  )) {
            strncpy(options->pgm->id, "#pGoblin" VERS, PGOBLIN_STRING_ID_LEN);
            options->pgm->maxx = PGOBLIN_PROGSIZE;
            options->pgm->curr = 0;
        } else {
            errno = ENOMEM;
    }   }
    ifBLIN_QV3(flags) fprintf(stderr, "-pgoblin_init %s\n", options ? "OK" : "FAIL");
    return(options);
}

#define R options->io[r]

static void
dump_ioreg(pgoblin_main *options, int r) {
    u_int32_t wdr, cdr;

    fprintf( stderr, "[%c](%06X%s%s%s%s%s%s%s%s%s%s%s%s):"
           , pgoblin_regn[r], R.flags
           , (R.flags & PGOBLIN_PATHPARM) ? " path" : ""
           , (R.flags & PGOBLIN_FREEPATH) ? "*"     : ""
           , (R.flags & PGOBLIN_FILEOVER) ? " over" : ""
           , (R.flags & PGOBLIN_TEXTPARM) ? " text" : ""
           , (R.flags & PGOBLIN_FREETEXT) ? "*"     : ""
           , (R.flags & PGOBLIN_BINPARM ) ? " bin"  : ""
           , (R.flags & PGOBLIN_MIFEDES ) ? " mife" : ""
           , (R.flags & PGOBLIN_FREEMIFE) ? "*"     : ""
           , (R.flags & PGOBLIN_PQRESULT) ? " pq"   : ""
           , (R.flags & PGOBLIN_FREEPQRE) ? "*"     : ""
           , (R.flags & PGOBLIN_OUTSET  ) ? " out"  : ""
           , (R.flags & PGOBLIN_FREEOUTS) ? "*"     : ""
           );
    if  (R.mife) {
        fprintf(stderr, " mife=%08X", (u_int32_t)R.mife);
        fprintf(stderr, " pid=%d", R.pid);
    }
    if  (R.pq) fprintf(stderr, " pq=%08X", (u_int32_t)R.pq);

    wdr = (u_int32_t)R.wri;
    if  (wdr == (u_int32_t)write) fprintf(stderr, " wri=write");
    else if (wdr == (u_int32_t)mife_writ) fprintf(stderr, " wri=mife_writ");
    else if (wdr) fprintf(stderr, " wri=%08X", wdr);

    cdr = (u_int32_t)R.clo;
    if  (cdr == (u_int32_t)close) fprintf(stderr, " clo=close");
    else if (cdr) fprintf(stderr, " clo=%08X", cdr);

    if  (wdr || cdr) fprintf(stderr, " onu=%d", R.onu);
    fprintf(stderr, "\n");

    if  (R.path) fprintf(stderr, " path=%s~\n", R.path);
    if  (R.text) {
        if  (!(R.flags & PGOBLIN_BINPARM)) {
            fprintf(stderr, " text=%s~\n", (char*)R.text);
        } else ifBLIN_QV5(options->flags) {
            int i;

            fprintf(stderr, " text=");
            for (i = 0; i < R.length; i++) {
                if  (  ((((u_char*)R.text)[i] >= 0177) && (((u_char*)R.text)[i] < 0300))
                    || (((u_char*)R.text)[i] < ' ')
                    ) {
                    fprintf(stderr, "\\%3o", ((u_char*)R.text)[i]);
                } else if (((char*)R.text)[i] == '\\') {
                    fprintf(stderr, "\\");
                } else {
                    fprintf(stderr, "%c", ((char*)R.text)[i]);
            }   }
            fprintf(stderr, "~\n");
        } else {
            fprintf(stderr, " text %08X length %d\n", (u_int32_t)R.text, R.length);
}   }   }

static void
iorassert(pgoblin_main *options, pgoblin_command cmd) {
    int r;
    u_int32_t e;
    u_char rr[PGOBLIN_REGSIZE + 1], rp[PGOBLIN_REGSIZE + 1];
    const u_char hex[16] = "0123456789ABCDEF";
#   define W "cmd[%d]=%03X fl[%d]=%08X "

    for (r = 0; r < PGOBLIN_REGSIZE; r++) {
        e = 0;
        if  (R.path) e |= PGOBLIN_PATHPARM;
          else if (R.flags & (PGOBLIN_PATHPARM | PGOBLIN_FREEPATH))
            warnx(W "  path", options->ppoint, cmd, r, R.flags);
        if  (R.text) e |= PGOBLIN_TEXTPARM;
          else if (R.flags & (PGOBLIN_TEXTPARM | PGOBLIN_FREETEXT))
            warnx(W "  text", options->ppoint, cmd, r, R.flags);
        if  (R.mife) e |= PGOBLIN_MIFEDES;
          else if (R.flags & (PGOBLIN_MIFEDES | PGOBLIN_FREEMIFE))
            warnx(W "  mife", options->ppoint, cmd, r, R.flags);
        if  (R.pq) e |= PGOBLIN_PQRESULT;
          else if (R.flags & (PGOBLIN_PQRESULT | PGOBLIN_FREEPQRE))
            warnx(W "  pq", options->ppoint, cmd, r, R.flags);
        ifBLIN_QV2(options->flags)
            if  (!(R.flags & PGOBLIN_ANYIVL) && e)
                warnx(W "  %02X", options->ppoint, cmd, r, R.flags, e);
        if  (R.wri) {
            e |= PGOBLIN_OUTSET;
            ifBLIN_QV2(options->flags)
                if  (!(R.flags & PGOBLIN_OUTSET))
                    warnx(W "  wri", options->ppoint, cmd, r, R.flags);
        } else if (R.flags & (PGOBLIN_OUTSET | PGOBLIN_FREEOUTS)) {
            warnx(W "  wri", options->ppoint, cmd, r, R.flags);
        }
        ifBLIN_QV2(options->flags)
            if  (R.clo && !(R.flags & PGOBLIN_FREEOUTS))
                warnx(W "  clo", options->ppoint, cmd, r, R.flags);
        rr[r] = e;
    }
    ifBLIN_QV3(options->flags) {
        /*    - ,
         *     ,
         *     .
         */
        for (r = 0; r < PGOBLIN_REGSIZE; r++) {
            rp[r] = pgoblin_regn[rr[r]];
            if  (rr[r] == (R.flags & (PGOBLIN_ANYIVL | PGOBLIN_OUTSET))) rr[r] = '=';
              else rr[r] = pgoblin_regn[R.flags & (PGOBLIN_ANYIVL | PGOBLIN_OUTSET)];
        }
        rr[PGOBLIN_REGSIZE] = rp[PGOBLIN_REGSIZE] = '\0';
        fprintf(stderr, "regs %s\n", rp);
        fprintf(stderr, "flag %s\n", rr);
    }
    ifBLIN_QV5(options->flags) {
        /*    -    */
        fprintf(stderr, "      00     08     10     18     20     28     30     38\n");
        for (r = 0; r < PGOBLIN_REGSIZE; r++) rr[r] = ' ';
        rr[1] = '0';
        rr[56] = '\0';
        for (r = 0; r < PGOBLIN_REGSIZE; r++) {
            e = ((r << 3) & 070) | ((r >> 3) & 007);
            rr[(r & 007) * 7 + 6] = hex[(e >> 4) & 0x0F];
            rr[(r & 007) * 7 + 7] = hex[e & 0x0F];
            rr[(r & 007) * 7 + 9] = pgoblin_regn[e];
            if  ((r & 007) == 007) {
                rr[2] = hex[(r >> 3) & 007];
                fprintf(stderr, "%s\n", rr);
}   }   }   }
#undef W
#undef R

static int
/**********************************************************************
 **********************************************************************
 **                                                                  **/
cleareg(pgoblin_main *options, pgoblin_io *r, u_int32_t m) {       /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
    int ex = EX_OK;

    ifBLIN_QV5(options->flags) fprintf(stderr, "+cleareg %08X %08X\n", (u_int32_t)r, m);
    if  (m & PGOBLIN_PATHPARM) {
        if  (r->path && (r->flags & PGOBLIN_FREEPATH)) free(r->path);
        r->path = NULL;
        r->flags &= ~(PGOBLIN_FREEPATH | PGOBLIN_FILEOVER | PGOBLIN_PATHPARM);
    }
    if  (m & PGOBLIN_TEXTPARM) {
        if  (r->text && (r->flags & PGOBLIN_FREETEXT)) free(r->text);
        r->text = NULL;
        r->length = 0;
        r->flags &= ~(PGOBLIN_FREETEXT | PGOBLIN_TEXTPARM | PGOBLIN_BINPARM);
    }
    if  (m & PGOBLIN_PQRESULT) {
        if  (r->pq && (r->flags & PGOBLIN_FREEPQRE)) PQclear(r->pq);
        r->pq = NULL;
        r->flags &= ~(PGOBLIN_FREEPQRE | PGOBLIN_PQRESULT);
    }
    if  (m & PGOBLIN_MIFEDES) {
        if  (r->mife && (r->flags & PGOBLIN_FREEMIFE) && (ex = mife_close(r->mife))) {
            ifBLIN_QV1(options->flags) warn("mife_close in cleareg");
        }
        r->mife = NULL;
        r->offset = 0;
        r->pid = 0;
        r->flags &= ~(PGOBLIN_FREEMIFE | PGOBLIN_MIFEDES);
    }
    if  (m & PGOBLIN_OUTSET) {
        if  (r->clo && (r->flags & PGOBLIN_FREEOUTS) && (ex = r->clo(r->onu))) {
            ifBLIN_QV1(options->flags) warn("clo in cleareg");
        }
        r->clo = NULL;
        r->wri = NULL;
        r->onu = 0;
        r->flags &= ~(PGOBLIN_OUTSET | PGOBLIN_FREEOUTS);
    }
    ifBLIN_QV5(options->flags) fprintf(stderr, "-cleareg %d\n", ex);
    return(ex);
}

#define Cx PGOBLIN_IGNORE   /* impossible         */
#define Co PGOBLIN_OPEN     /* open               */
#define Cf PGOBLIN_PATHPARM /* free path          */
#define Cw PGOBLIN_OUTSET   /* free wri           */
#define Ct PGOBLIN_TEXTPARM /* free text          */
#define Cq PGOBLIN_PQRESULT /* free pq            */
#define Cc PGOBLIN_MIFEDES  /* free mife          */
#define Ce PGOBLIN_FORKOUT  /* fork               */
#define Cs PGOBLIN_FREEPATH /* stream file out    */
#define Cp PGOBLIN_FREETEXT /* copy text out      */
#define Cu PGOBLIN_FREEPQRE /* unstore pq out     */

static const u_int32_t score[32][4] =
 /*ftmqw          Pf           Pm           Pt           Pq   Wish*/
{/*00000*/{0              ,0           ,          Ce,0           }
,/*10000*/{ Co|Cf|      Cs, Co|      Cp, Co         , Co|      Cu}
,/*01000*/{       Ct      ,    Ct      ,    Ct|   Ce,    Ct      }
,/*11000*/{ Co|Cf|Ct|   Cs, Co|Ct|   Cp, Co|Ct      , Co|Ct|   Cu}
,/*00100*/{ Cc            , Cc         ,          Ce, Cc         }
,/*10100*/{ Co|Cf|      Cs, Co|      Cp, Co         , Co|      Cu}
,/*01100*/{ Cc|   Ct      , Cc|Ct      ,    Ct|   Ce, Cc|Ct      }
,/*11100*/{ Co|Cf|Ct|   Cs, Co|Ct|   Cp, Co|Ct      , Co|Ct|   Cu}
,/*00010*/{          Cq   ,       Cq   ,       Cq|Ce,       Cq   }
,/*10010*/{ Co|Cf|   Cq|Cs, Co|   Cq|Cp, Co|   Cq   , Co|   Cq|Cu}
,/*01010*/{       Ct|Cq   ,    Ct|Cq   ,    Ct|Cq|Ce,    Ct|Cq   }
,/*11010*/{ Co|Cf|Ct|Cq|Cs, Co|Ct|Cq|Cp, Co|Ct|Cq   , Co|Ct|Cq|Cu}
,/*00110*/{ Cc|      Cq   , Cc|   Cq   ,       Cq|Ce, Cc|   Cq   }
,/*10110*/{ Co|Cf|   Cq|Cs, Co|   Cq|Cp, Co|   Cq   , Co|   Cq|Cu}
,/*01110*/{ Cc|   Ct|Cq   , Cc|Ct|Cq   ,    Ct|Cq|Ce, Cc|Ct|Cq   }
,/*11110*/{ Co|Cf|Ct|Cq|Cs, Co|Ct|Cq|Cp, Co|Ct|Cq   , Co|Ct|Cq|Cu}
,/*00001*/{             Cs,          Cp,0           ,          Cu}
,/*10001*/{    Cf|      Cs,          Cp,0           ,          Cu}
,/*01001*/{       Ct|   Cs,    Ct|   Cp,    Ct      ,    Ct|   Cu}
,/*11001*/{    Cf|Ct|   Cs,    Ct|   Cp,    Ct      ,    Ct|   Cu}
,/*00101*/{             Cs,          Cp,0           ,          Cu} /* IO_RG0 */
,/*10101*/{    Cf|      Cs,          Cp,0           ,          Cu}
,/*01101*/{       Ct|   Cs,    Ct|   Cp,    Ct      ,    Ct|   Cu}
,/*11101*/{    Cf|Ct|   Cs,    Ct|   Cp,    Ct      ,    Ct|   Cu}
,/*00011*/{          Cq|Cs,       Cq|Cp,       Cq   ,       Cq|Cu}
,/*10011*/{    Cf|   Cq|Cs,       Cq|Cp,       Cq   ,       Cq|Cu}
,/*01011*/{       Ct|Cq|Cs,    Ct|Cq|Cp,    Ct|Cq   ,    Ct|Cq|Cu}
,/*11011*/{    Cf|Ct|Cq|Cs,    Ct|Cq|Cp,    Ct|Cq   ,    Ct|Cq|Cu}
,/*00111*/{          Cq|Cs,       Cq|Cp,       Cq   ,       Cq|Cu}
,/*10111*/{    Cf|   Cq|Cs,       Cq|Cp,       Cq   ,       Cq|Cu}
,/*01111*/{       Ct|Cq|Cs,    Ct|Cq|Cp,    Ct|Cq   ,    Ct|Cq|Cu}
,/*11111*/{    Cf|Ct|Cq|Cs,    Ct|Cq|Cp,    Ct|Cq   ,    Ct|Cq|Cu}
}/* Reg */;

static u_int32_t
/**********************************************************************
 **********************************************************************
 **                                                                  **/
procomp(pgoblin_main *options, pgoblin_io *r, u_int32_t ch) {      /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
    int i, j;
    u_int32_t price, abil;

    abil = 0;
    for (j = 0; j < 5; j++) { /* From */
        if  (  ((j == 0) && (r->path))
            || ((j == 1) && (r->text))
            || ((j == 2) && (r->mife))
            || ((j == 3) && (r->pq  ))
            || ((j == 4) && (r->wri ))
            ) {
            abil |= (1 << j);
    }   }

    price = 0;
    for (i = 0; i < 4; i++) { /* To */
        if  (  ((i == 0) && (ch & PGOBLIN_PATHPARM))
            || ((i == 1) && (ch & PGOBLIN_TEXTPARM))
            || ((i == 2) && (ch & PGOBLIN_MIFEDES ))
            || ((i == 3) && (ch & PGOBLIN_PQRESULT))
            ) {
            ifBLIN_QV4(options->flags) fprintf(stderr, "%d abil=%02X %04X\n", i, abil, score[abil][i]);
            price |= score[abil][i];
    }   }
    ifBLIN_QV4(options->flags) fprintf(stderr, "abil=%02X %04X\n", abil, price);
    return(price);
}

static int
/**********************************************************************
 **********************************************************************
 **                                                                  **/
propush(pgoblin_main *options, pgoblin_io *r, u_int32_t ch) {       /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
    u_int32_t price;
    int ex = EX_OK;

    ifBLIN_QV5(options->flags) fprintf( stderr, "+propush ch=%02X io=%08X%s%s%s%s\n"
                                      , ch, r->flags
                                      , r->path ? " path" : ""
                                      , r->text ? " text" : ""
                                      , r->wri  ? " mife" : ""
                                      , r->pq   ? " pq"   : ""
                                      );
    if  (!(ch & PGOBLIN_OUTSET)) {
        ifBLIN_QV1(options->flags) warnx("No output in propush");
        ERROUT(EX_DATAERR, EINVAL);
    } else if (  ((r->flags & PGOBLIN_PATHPARM) && !r->path)
              || ((r->flags & PGOBLIN_TEXTPARM) && !r->text)
              || ((r->flags & PGOBLIN_OUTSET  ) && !r->wri )
              || ((r->flags & PGOBLIN_PQRESULT) && !r->pq  )
              ) {
        ifBLIN_QV1(options->flags) warnx("No info f=%08X", r->flags);
        ERROUT(EX_DATAERR, EINVAL);
    }
    price = procomp(options, r, ch);
    if  (price & Cx) {
        ifBLIN_QV1(options->flags) warnx("No way @%d f=%08X", options->ppoint, r->flags);
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  (price & Co) {
        r->onu = open( r->path
                     , O_WRONLY | O_CREAT | ((r->flags & PGOBLIN_FILEOVER) ? O_TRUNC : O_APPEND)
                     , 0666
                     );
        if  (r->onu < 0) {
            r->onu = 0;
            ex = EX_IOERR;
            goto out;
        }
        r->clo = close;
        r->wri = mife_writ;
        r->flags |= PGOBLIN_FREEOUTS;
    }
    ex = cleareg(options, r, price);
    if  (!ex) options->exeflags |= price;
out:
    ifBLIN_QV5(options->flags) fprintf(stderr, "-propush %d\n", ex);
    return(ex);
}

#define DC 0x000F /* price              */
#define Dx 0x8000 /* impossible         */
#define Dq 0x4000 /* read to text       */
#define Do 0x2000 /* open file name     */
#define Dr 0x1000 /* read file          */
#define Dn 0x0800 /* temp file name     */
#define Dw 0x0400 /* write file         */
#define Dm 0x0200 /* do mife from text  */
#define Da 0x0100 /* throw pq to file   */
#define Db 0x0080 /* throw mife to file */

static const u_int16_t dcore[4][4] =
      /*         Pf          Pt          Pm       Pq   Reg */
{/* Pf */{          0, Dx|Dn|Dw|7, Dx|Dn|Db|9, Dx|Dn|Da|8}
,/* Pt */{    Do|Dr|3,          0,    Dr   |2,    Dq   |5}
,/* Pm */{    Do   |1,       Dm|1,          0,    Dq|Dm|6}
,/* Pq */{ Dx     |15, Dx     |15, Dx     |15,          0}
}/*Wish*/;

static u_int32_t
/**********************************************************************
 **********************************************************************
 **                                                                  **/
coercomp(pgoblin_main *options, pgoblin_io *r, u_int32_t ch) {      /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
    int i, j;
    u_int32_t price;

    price = Dx | 15;
    for (i = 0; i < 4; i++) { /* To */
        if  (  ((i == 0) && (ch & PGOBLIN_PATHPARM))
            || ((i == 1) && (ch & PGOBLIN_TEXTPARM))
            || ((i == 2) && (ch & PGOBLIN_MIFEDES ))
            || ((i == 3) && (ch & PGOBLIN_PQRESULT))
            ) {
            for (j = 0; j < 4; j++) { /* From */
                if  (  (  ((j == 0) && (r->path))
                       || ((j == 1) && (r->text))
                       || ((j == 2) && (r->mife))
                       || ((j == 3) && (r->pq  ))
                       )
                    && (dcore[i][j] & DC) < (price & DC)
                    ) {
                    price = dcore[i][j];
                    ifBLIN_QV4(options->flags)
                        fprintf(stderr, "i=%d j=%d %04X\n", i, j, dcore[i][j]);
    }   }   }   }
    return(price);
}

static int
/**********************************************************************
 **********************************************************************
 **                                                                  **/
coercion(pgoblin_main *options, pgoblin_io *r, u_int32_t ch) {      /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
    u_int32_t price;
    int ex = EX_OK;

    ifBLIN_QV5(options->flags) fprintf( stderr, "+coercion ch=%02X io=%08X%s%s%s%s\n"
                                      , ch, r->flags
                                      , r->text ? " path" : ""
                                      , r->text ? " text" : ""
                                      , r->mife ? " mife" : ""
                                      , r->pq   ? " pq"   : ""
                                      );
    if  (ch & PGOBLIN_OUTSET) {
        ifBLIN_QV1(options->flags) warnx("Output in coercion");
        ERROUT(EX_DATAERR, EINVAL);
    } else if (  ((r->flags & PGOBLIN_PATHPARM) && !r->path)
              || ((r->flags & PGOBLIN_TEXTPARM) && !r->text)
              || ((r->flags & PGOBLIN_MIFEDES ) && !r->mife)
              || ((r->flags & PGOBLIN_PQRESULT) && !r->pq  )
              ) {
        ifBLIN_QV1(options->flags) warnx("No info f=%08X", r->flags);
        ERROUT(EX_DATAERR, EINVAL);
    }
    price = coercomp(options, r, ch);
    if  (price & Dx) {
        ifBLIN_QV1(options->flags) warnx("No way @%d f=%08X", options->ppoint, r->flags);
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  (price & Dq) {
        ssize_t k, c, e, l;
        char *s;

        for (k = 0, e = 0; k < PQntuples(r->pq) && e >= 0; k++) {
            for (c = 0; c < PQnfields(r->pq) && e >= 0; c++) {
                if  (!PQgetisnull(r->pq, k, c))
                    e += PQgetlength(r->pq, k, c);
        }   }
        if  (!(s = malloc(e + 1))) {
            ifBLIN_QV1(options->flags) warnx("No mem");
            ERROUT(EX_OSERR, ENOMEM);
        }
        r->text = s;
        r->length = e;
        r->flags |= PGOBLIN_FREETEXT | PGOBLIN_BINPARM;
        ifBLIN_QV4(options->flags)
            fprintf(stderr, "text=%08X length=%d\n", (u_int32_t)r->text, r->length);
        for (k = 0, e = 0; k < PQntuples(r->pq) && e >= 0; k++) {
            for (c = 0; c < PQnfields(r->pq) && e >= 0; c++) {
                if  (!PQgetisnull(r->pq, k, c)) {
                    l = PQgetlength(r->pq, k, c);
                    bcopy(PQgetvalue(r->pq, k, c), s, l);
                    s += l;
        }   }   }
        *s = '\0';
    }
    if  (price & Do) {
        r->offset = 0;
        r->pid = 0;
        r->mife = mife_open(PGOBLIN_MIFEFLAGS, r->path);
        if  (!r->mife) {
            ifBLIN_QV1(options->flags) warn("coercion mife open %s", r->path);
            ex = EX_IOERR;
            goto out;
    }   }
    if  (price & Dr) {
        r->length = mife_read(r->mife, 0, 0);
        if  (!(r->length)) {
            ifBLIN_QV1(options->flags) warnx("mife_read");
            ex = EX_IOERR;
            goto out;
        }
        r->text = mife_get(r->mife, 0);
        if  (!(r->text)) {
            ifBLIN_QV1(options->flags) warnx("mife_get");
            ex = EX_IOERR;
            goto out;
    }   }
    if  (price & Dn) /* XXXX */;
    if  (price & Dw) /* XXXX */;
    if  (price & Dm) {
        r->flags |= PGOBLIN_FREEMIFE;
        r->offset = 0;
        r->pid = 0;
        if  ((r->flags & PGOBLIN_BINPARM) && !r->length) r->length = strlen(r->text);
        if  (r->flags & PGOBLIN_FREETEXT) {
            r->mife = mife_opem( (r->flags & PGOBLIN_BINPARM) ? 0 : PGOBLIN_MIFEFLAGS
                               , r->text
                               , r->length
                               );
        } else {
            r->mife = mife_opes(PGOBLIN_MIFEFLAGS, r->text, r->length);
        }
        if  (!r->mife) {
            ifBLIN_QV1(options->flags) warnx("mife_ope{m|s}");
            ex = EX_IOERR;
            goto out;
    }   }
    if  (price & Da) /* XXXX */;
    if  (price & Db) /* XXXX */;
out:
    ifBLIN_QV5(options->flags) fprintf(stderr, "-coercion %d\n", ex);
    return(ex);
}

static int
pgoblin_inx(pgoblin_main *options, pgoblin_io *r, int p[2], int c) {
    int ex = EX_OK; /* Errors must be < 0 */

    if  (c) close(p[1]);
    if  (!(r->mife = mife_opef(PGOBLIN_MIFEFLAGS, p[0]))) {
        ifBLIN_QV1(options->flags) warn("mife_opef inx");
        ex = -EX_NOINPUT;
    } else r->flags |= PGOBLIN_MIFEDES | PGOBLIN_FREEMIFE;
    r->pid = 0;
    r->offset = 0;
    return(ex);
}

static int
pgoblin_outx(pgoblin_main *options, pgoblin_io *r, int p[2], int c) {
    int ex = EX_OK; /* Errors must be < 0 */

    if  (c) close(p[0]);
    r->onu = p[1];
    r->wri = mife_writ;
    r->clo = close;
    r->flags |= PGOBLIN_OUTSET | PGOBLIN_FREEOUTS;
    return(ex);
}

static int
internal_fork(pgoblin_main *options, pgoblin_r *r, char **argv) {
    int ex = EX_OK;

    ifBLIN_QV5(options->flags)
        fprintf( stderr, "+internal_fork %s %d %d %d %d %d %d argv %s\n"
               , pgoblin.a[PGOBLIN_fork].str
               , r[PGO_COUT].r, r[PGO_CCTL].r, r[PGO_CIN].r
               , r[PGO_CCON].r, r[PGO_CJOB].r, r[PGO_CSTY].r
               , argv ? "OK" : "NULL"
               );
    if  ((ex = fork()) < 0) {
        ifBLIN_QV1(options->flags) warn("fork #1");
    } else if (ex == 0) {
        /*********************************
         * ,    *
         *********************************/
    ;   ex = pgoblin_connown(options);
    ;   options->exeflags |= PGOBLIN_FORKED;
    ;   if  (options->exeflags & PGOBLIN_FORKCTL) {
    ;       ifBLIN_QV1(options->flags) warnx("fork with CTL");
    ;       ex = -EX_SOFTWARE;
    ;       exit(ex);
    ;   }
    ;   if  (options->exeflags & PGOBLIN_FORKIN) {
    ;       close(options->pip2[1]);
    ;       if  (!argv) {
    ;           if  ((ex = pgoblin_inx(options, IO_IN, options->pip2, 0))) exit(ex);
    ;       } else {
    ;           if  (dup2(options->pip2[0], fileno(stdin)) < 0) {
    ;               ifBLIN_QV1(options->flags) warn("dup2 #2 in pid=%d", getpid());
    ;               ex = -EX_OSERR;
    ;               exit(ex);
    ;           }
    ;           if  (options->pip2[0] != fileno(stdin)) close(options->pip2[0]);
    ;       }
    ;   } else if (argv && (dup2(IO_IN->mife->fd, fileno(stdin)) < 0)) {
    ;       ifBLIN_QV1(options->flags) warn("dup2 #4");
    ;       ex = -EX_OSERR;
    ;       exit(ex);
    ;   }
    ;   if  (options->exeflags & PGOBLIN_FORKOUT) {
    ;       if  (!argv) {
    ;           if  ((ex = pgoblin_outx(options, IO_OUT, options->pip0, 1))) exit(ex);
    ;       } else {
    ;           close(options->pip0[0]);
    ;           if  (dup2(options->pip0[1], fileno(stdout)) < 0) {
    ;               ifBLIN_QV1(options->flags) warn("dup2 #1");
    ;               ex = -EX_OSERR;
    ;               exit(ex);
    ;           }
    ;           if  (options->pip0[1] != fileno(stdout)) close(options->pip0[1]);
    ;       }
    ;   } else if (argv && (dup2(IO_OUT->onu, fileno(stdout)) < 0)) {
    ;       ifBLIN_QV1(options->flags) warn("dup2 #3");
    ;       ex = -EX_OSERR;
    ;       exit(ex);
    ;   }
    } else {
        pgoblin_job *j;

        j = mular_add(*r[PGO_CJOB].g.j);
        j->cmdn = options->ppoint;
        j->pid = ex;
        j->status = PGOBLIN_NOSTATUS;
    }
    ifBLIN_QV5(options->flags) fprintf(stderr, "-internal_fork %d\n", ex);
    return(ex);
}

int
pgoblin_fork(pgoblin_main *options, pgoblin_r *r) {
    int ex = EX_OK;

    ifBLIN_QV4(options->flags)
        fprintf( stderr, "+pgoblin_fork %s %d %d %d %d %d %d\n"
               , pgoblin.a[PGOBLIN_fork].str
               , r[PGO_COUT].r, r[PGO_CCTL].r, r[PGO_CIN].r
               , r[PGO_CCON].r, r[PGO_CJOB].r, r[PGO_CSTY].r
               );
    if  ((ex = internal_fork(options, r, NULL)) > 0) {
        if  ((ex >= 0) && options->exeflags & PGOBLIN_FORKIN) {
            ex = pgoblin_outx(options, IO_IN, options->pip2, 1);
        }
        if  ((ex >= 0) && (options->exeflags & PGOBLIN_FORKCTL)) {
            ex = pgoblin_outx(options, IO_CTL, options->pip1, 1);
        }
        if  ((ex >= 0) && (options->exeflags & PGOBLIN_FORKOUT)) {
            ex = pgoblin_inx(options, IO_OUT, options->pip0, 1);
        }
        options->lexlevel++;
        ex = 0;
    }
    ifBLIN_QV4(options->flags) fprintf(stderr, "-pgoblin_fork %d\n", ex);
    return(ex);
}

int
pgoblin_exec(pgoblin_main *options, pgoblin_r *r) {
    int ex = EX_OK;
    int t, n, c, g = 0, i;
    char **argv;

    ifBLIN_QV4(options->flags)
        fprintf(stderr, "+pgoblin_exec %d %d %d %d\n", r[0].r, r[1].r, r[3].r, r[2].r);
    n = 1;
    if  (r[PGO_CCTL].r) {
        n = PQntuples(IO_CTL->pq);
        g = PQnfields(IO_CTL->pq);
    }
    for (t = 0; (t < n) && !ex; t++) {
        if  (r[PGO_CCTL].r) {
            i = 0;
            if  (!(argv = malloc((g + 1) * sizeof(char**)))) {
                ex = EX_OSERR;
                break;
            }
            for (c = 0; c < g; c++) {
                if  (!PQgetisnull(IO_CTL->pq, t, c)) argv[i++] = PQgetvalue(IO_CTL->pq, t, c);
            }
        } else {
            for (c = 0; ; c++) {
                i = c;
                if  (c && options->pgm->code[options->ppoint + c].cmd != PGOBLIN_cont) break;
                if  (!options->pgm->code[options->ppoint + c].e.lit) break;
            }
            if  (!(argv = malloc((i + 1) * sizeof(char**)))) {
                ex = EX_OSERR;
                break;
            }
            for (c = 0; c < i; c++) argv[c] = options->pgm->code[options->ppoint + c].e.lit;
        }
        if  (!i) continue;
        argv[i] = NULL;
        ex = internal_fork(options, r, argv);
        if  (ex == 0) {
            int g;

            for (g = 0; g < PGOBLIN_REGSIZE; g++) {
#               define R options->io[g]

                if  (R.mife && (R.mife->fd != fileno(stdin)) && (R.mife->fd != fileno(stdout))) {
                    if  ((R.flags & PGOBLIN_FREEMIFE) && (ex = mife_close(R.mife)))
                        ifBLIN_QV1(options->flags) warn("mife_close in exec");
                }
                if  (R.clo && (R.onu != fileno(stdin)) && (R.onu != fileno(stdout))) {
                    if  ((R.flags & PGOBLIN_FREEOUTS) && (ex = R.clo(R.onu)))
                        ifBLIN_QV1(options->flags) warn("clo in exec");
            }   }
            if  (ex < 0) /* XXXX */;
            execve(argv[0], argv, options->envp);
            ex = -EX_OSERR;
            ifBLIN_QV1(options->flags) warn("execve");
        } else if (ex > 0) {
            ex = 0;
    }   }
    if  (!ex) {
        if  (options->exeflags & PGOBLIN_FORKIN) {
            ex = pgoblin_outx(options, IO_IN, options->pip2, 1);
        }
        if  (!ex && (options->exeflags & PGOBLIN_FORKCTL)) {
            ex = pgoblin_outx(options, IO_CTL, options->pip1, 1);
        }
        if  (!ex && (options->exeflags & PGOBLIN_FORKOUT)) {
            ex = pgoblin_inx(options, IO_OUT, options->pip0, 1);
    }   }
    ifBLIN_QV4(options->flags) fprintf( stderr, "-pgoblin_exec %d\n", ex);
    return(ex);
}

int
pgoblin_pipe(pgoblin_main *options, pgoblin_r *r) {
    int ex = EX_OK;
    int pipa[2];

    ifBLIN_QV4(options->flags) fprintf(stderr, "+pgoblin_pipe %d %d\n", r[PGO_COUT].r, r[PGO_CIN].r);
    if  (pipe(pipa) < 0) {
        ifBLIN_QV1(options->flags) warn("pipe in pipe");
        close(pipa[0]);
        close(pipa[1]);
        ex = EX_OSERR;
    }
    if  (!(ex = pgoblin_inx(options, IO_IN, pipa, 0))) ex = pgoblin_outx(options, IO_OUT, pipa, 0);
    ifBLIN_QV4(options->flags) fprintf(stderr, "-pgoblin_pipe %d\n", ex);
    return(ex);
}

int
pgoblin_return(pgoblin_main *options, pgoblin_r *r) {
    int ex = EX_OK;
    u_int32_t m;
    ssize_t g;

    ifBLIN_QV4(options->flags) fprintf(stderr, "+pgoblin_return\n");
    if  (options->exeflags & PGOBLIN_END || !(options->exeflags & PGOBLIN_INTRAP)) {
        for (g = 0; g < PGOBLIN_REGSIZE; g++) {
            m = PGOBLIN_PATHPARM | PGOBLIN_TEXTPARM | PGOBLIN_PQRESULT;
            if  (g) m |= PGOBLIN_OUTSET | PGOBLIN_MIFEDES;
            cleareg(options, (&options->io[g]), m);
            if  (options->conn[g].conn && (options->conn[g].flags & PGOBLIN_CLOSECONN)) {
                PQfinish(options->conn[g].conn);
                options->conn[g].conn = NULL;
            }
            if  (options->job[g]) {
                mular_destroy(options->job[g]);
                options->job[g] = NULL;
            }
            if  (options->style[g]) options->style[g] = NULL;
        }
        if  (options->exeflags & PGOBLIN_FORKED) exit(ex);
        options->exeflags |= PGOBLIN_END;
    } else {
        options->ppoint = options->rtrap;
        options->exeflags &= ~PGOBLIN_INTRAP;
    }
    ifBLIN_QV4(options->flags) fprintf(stderr, "-pgoblin_return %d\n", ex);
    return(ex);
}

int
pgoblin_exit(pgoblin_main *options, pgoblin_r *r) {
    int ex;
    ifBLIN_QV4(options->flags) fprintf(stderr, "+pgoblin_exit\n");
    ex = pgoblin_return(options, r);
    ifBLIN_QV4(options->flags) fprintf(stderr, "-pgoblin_exit %d\n", ex);
    exit(ex);
}

#define CICLEWARN(N,A) { int CICLEWARN_INTERNAL = N;                                                 \
    ifBLIN_QV2(options->flags)                                                                       \
        warnx( "cmd[%d]=[%03X(%s %c%c%c%c%c%c)].fl=%08X %s ch=%02X " A                               \
             , options->ppoint, cmd, a_cmd->str                                                      \
             , pgoblin_regn[r[0].r], pgoblin_regn[r[1].r], pgoblin_regn[r[2].r]                      \
             , pgoblin_regn[r[3].r], pgoblin_regn[r[4].r], pgoblin_regn[r[5].r]                      \
             , a_cmd->flags, pgoblin.p[CICLEWARN_INTERNAL].name, r[CICLEWARN_INTERNAL].c             \
             );                                                                                      \
    continue;                                                                                        \
}
#define CICLEERR(A) {                                                                                \
    ifBLIN_QV1(options->flags)                                                                       \
        warn( "cmd[%d]=%03X(%s %c%c%c%c%c%c) %08X %02X %02X %02X %02X %02X %02X " A                  \
            , options->ppoint, cmd, a_cmd->str                                                       \
            , pgoblin_regn[r[0].r], pgoblin_regn[r[1].r], pgoblin_regn[r[2].r]                       \
            , pgoblin_regn[r[3].r], pgoblin_regn[r[4].r], pgoblin_regn[r[5].r]                       \
            , a_cmd->flags, r[0].c, r[1].c, r[2].c, r[3].c, r[4].c, r[5].c                           \
            );                                                                                       \
}
#define DUMPIOREGS { int DUMPIOREGS_INTERNAL;                                                        \
    for (DUMPIOREGS_INTERNAL = 0; DUMPIOREGS_INTERNAL < SYMREGS; DUMPIOREGS_INTERNAL++) {            \
        if (  (  (DUMPIOREGS_INTERNAL == PGO_COUT)                                                   \
              || (DUMPIOREGS_INTERNAL == PGO_CCTL)                                                   \
              || (DUMPIOREGS_INTERNAL == PGO_CIN)                                                    \
              )                                                                                      \
           && !(r[DUMPIOREGS_INTERNAL].c & PGOBLIN_ILLEGAL)                                          \
           ) {                                                                                       \
           dump_ioreg(options, r[DUMPIOREGS_INTERNAL].r);                                            \
}   }   }

int
/**********************************************************************
 **********************************************************************
 **                                                                  **/
pgoblin_execute(pgoblin_main *options) {                             /**
 **                                                                  **
 **********************************************************************
 **********************************************************************/
    int ex = EX_OK, i, g;
    pgoblin_r r[PGO_MAXR];

    ifBLIN_QV3(options->flags) fprintf( stderr, "+pgoblin_execute pgm@%d:%d\n"
                                      , options->pgm->curr, options->pgm->maxx
                                      );
    ifBLIN_QV5(options->flags) for (g = 0; g < PGOBLIN_REGSIZE; g++) dump_ioreg(options, g);
    ex = 0;
    options->lexlevel = 0;
    options->exeflags = 0;
    for (options->ppoint = 0; !(options->exeflags & PGOBLIN_END); options->ppoint++) {
        u_int32_t t;
        pgoblin_command cmd;
        const struct pgoblin_syntax_a *a_cmd;
        char *lit;

    ;   options->exeflags &= PGOBLIN_EXTABLE;
    ;   if  (ex) {
    ;       if  (options->exeflags & PGOBLIN_INTRAP) {
    ;           ifBLIN_QV2(options->flags) warnx("Fail in trap PC %d", options->ppoint);
    ;           options->exeflags |= PGOBLIN_END | PGOBLIN_ERROR;
    ;       } else if (options->exeflags & PGOBLIN_TRAPDEF) {
    ;           ifBLIN_QV2(options->flags)
                    warnx("Trap from %d to %d", options->ppoint, options->ctrap);
    ;           options->rtrap = options->ppoint - 1;
    ;           options->ppoint = options->ctrap;
    ;           options->exeflags |= PGOBLIN_INTRAP;
    ;       } else {
    ;           ifBLIN_QV2(options->flags) warnx("Fail PC %d", options->ppoint);
    ;           options->exeflags |= PGOBLIN_END | PGOBLIN_ERROR;
    ;   }   }
    ;   if  ((options->ppoint < options->pgm->curr) && !(options->exeflags & PGOBLIN_END)) {
    ;       t = options->pgm->code[options->ppoint].cmd;
    ;       if  ((cmd = t & PGOBLIN_COMMAND) == PGOBLIN_extended) {
    ;           cmd = (options->pgm->code[options->ppoint].e.cmd & PGOBLIN_COMMAND) | PGOBLIN_EXTENDED;
    ;           lit = NULL;
    ;       } else {
    ;           lit = options->pgm->code[options->ppoint].e.lit;
    ;       }
    ;       if  (cmd >= PGOBLIN_max || !(~pgoblin.a[cmd].flags & PGOBLIN_ILLEGAL)) {
    ;           ex = EX_DATAERR;
    ;           ifBLIN_QV1(options->flags) warnx("cmd[%d]=%03X unknown", options->ppoint, cmd);
    ;           continue;
    ;       } else if (!pgoblin.a[cmd].flags) {
    ;           continue;
    ;       } else if (!pgoblin.a[cmd].car) {
    ;           ex = EX_DATAERR;
    ;           ifBLIN_QV2(options->flags)
                    warnx("cmd[%d]=%03X(%s) Undefined", options->ppoint, cmd, pgoblin.a[cmd].str);
    ;           continue;                                                                                        \
    ;       }
    ;       a_cmd = &(pgoblin.a[cmd]);
    ;       if  (options->lexlevel) {
    ;           if  (a_cmd->flags & PGOBLIN_LEXLVUP) options->lexlevel++;
    ;           if  (a_cmd->flags & PGOBLIN_LEXLDOWN) options->lexlevel--;
    ;           continue;
    ;       }
    ;   } else {
    ;       t = 0;
    ;       lit = NULL;
    ;       cmd = PGOBLIN_return;
    ;       a_cmd = &(pgoblin.a[cmd]);
    ;       options->exeflags |= PGOBLIN_END;
    ;   }
    ;   ifBLIN_QV3(options->flags) {
    ;       fprintf( stderr, "####N%d e%06X f%08X c%03X %s"
                   , options->ppoint, options->exeflags, a_cmd->flags, cmd, pgoblin.a[cmd].str
                   );
    ;       if  (lit) fprintf(stderr, "=%s~\n", lit);
              else fprintf(stderr, "\n");
    ;   }
    ;   for (i = 0; i < PGO_MAXR; i++) {
    ;       if  (a_cmd->flags & pgoblin.p[i].sym) {
                int j;

    ;           j = PREG(cmd, i);
    ;           r[i].c = a_cmd->ch[j];
    ;           r[i].r = (t >> pgoblin.binparams[j]) & PGOBLIN_ARGMASK;
    ;           ifBLIN_QV3(options->flags)
                    fprintf(stderr, "%d->%d[%d] (%02X%s%s%s%s%s%s%s%s) %s "
                           , j, i, r[i].r, r[i].c
                           , (r[i].c & PGOBLIN_STYLE   ) ? " style": ""
                           , (r[i].c & PGOBLIN_JOB     ) ? " job"  : ""
                           , (r[i].c & PGOBLIN_CONNECT ) ? " conn" : ""
                           , (r[i].c & PGOBLIN_OUTSET  ) ? " out"  : ""
                           , (r[i].c & PGOBLIN_PQRESULT) ? " pq"   : ""
                           , (r[i].c & PGOBLIN_MIFEDES ) ? " mife" : ""
                           , (r[i].c & PGOBLIN_TEXTPARM) ? " text" : ""
                           , (r[i].c & PGOBLIN_PATHPARM) ? " path" : ""
                           , pgoblin.p[i].name
                           );
    ;           if  (i == PGO_CSTY) {
    ;               r[i].g.s = &(options->style[r[i].r]);
    ;               ifBLIN_QV3(options->flags) fprintf(stderr, "%08X\n", (u_int32_t)(r[i].g.s));
    ;           } else if (i == PGO_CJOB) {
    ;               r[i].g.j = &(options->job[r[i].r]);
    ;               ifBLIN_QV3(options->flags) fprintf(stderr, "%08X\n", (u_int32_t)(r[i].g.j));
    ;           } else if (i == PGO_CCON) {
    ;               r[i].g.c = &(options->conn[r[i].r]);
    ;               ifBLIN_QV3(options->flags) fprintf(stderr, "%08X\n", (u_int32_t)(r[i].g.c));
    ;           } else {
    ;               r[i].g.io = &(options->io[r[i].r]);
    ;               ifBLIN_QV3(options->flags) fprintf(stderr, "%08X\n", (u_int32_t)(r[i].g.io));
    ;           }
    ;       } else {
    ;           r[i].c = PGOBLIN_ILLEGAL;
    ;           r[i].r = 0;
    ;           r[i].g.io = NULL;
    ;   }   }
    ;   ifBLIN_QV4(options->flags) DUMPIOREGS;
    ;   /**************************
    ;    *   OUT *
    ;    **************************/
    ;   if  (a_cmd->flags & PGOBLIN_OUTSET) {
    ;       if  (!(CH_OUT & PGOBLIN_OUTSET)) {
    ;           if  (r[PGO_COUT].r) {
    ;               if  ((ex = cleareg(options, IO_OUT, PGOBLIN_PATHPARM | PGOBLIN_OUTSET))) {
    ;                   CICLEWARN(0, "Not clear path|wri OUT");
    ;           }   }
    ;       } else if ((ex = propush(options, IO_OUT, CH_OUT))) {
    ;           CICLEWARN(0, "No propush");
    ;       } else if ((r[PGO_COUT].r == r[PGO_CCTL].r) || (r[PGO_COUT].r == r[PGO_CIN].r)) {
    ;               /* copy XXXX */
    ;   }   }
    ;   /**************************
    ;    *   CTL *
    ;    **************************/
    ;   if  (a_cmd->flags & PGOBLIN_CTL) {
    ;       if  (CH_CTL & PGOBLIN_OUTSET) {
    ;           ex = EX_SOFTWARE;
    ;           CICLEWARN(1, "CTL&out");
    ;       } else if (r[PGO_CCTL].r) {
    ;           if  (  (r[PGO_CCTL].r == r[PGO_COUT].r)
                    && (CH_CTL & CH_OUT & (PGOBLIN_PATHPARM | PGOBLIN_TEXTPARM | PGOBLIN_PQRESULT))
                    ) {
    ;               ifBLIN_QV2(options->flags)
                        warnx( "cmd[%d]=[%03X(%s)]    \n"
                             , options->ppoint, cmd, a_cmd->str
                             );
    ;           }
    ;           if  (!(IO_CTL->flags & PGOBLIN_ANYIVL)) {
    ;               if  (r[PGO_CCTL].r != r[PGO_CIN].r) options->exeflags |= PGOBLIN_FORKCTL;
    ;           } else if ((ex = coercion(options, IO_CTL, CH_CTL))) {
    ;               CICLEWARN(1, "No coercion CTL");
    ;           }
    ;       } else if ((CH_CTL & PGOBLIN_TEXTPARM)) {
    ;           if  ((ex = cleareg(options, IO_CTL, PGOBLIN_TEXTPARM))) {
    ;               CICLEWARN(1, "Not ready text CTL");
    ;           }
    ;           if  (!lit) {
    ;               ex = EX_SOFTWARE;
    ;               CICLEWARN(1, "No lit");
    ;           } else {
    ;               IO_CTL->text = lit;
    ;           }
    ;       } else if ((CH_CTL & PGOBLIN_PQRESULT)) {
    ;           /* #exec known to work this out */
    ;       } else if ((ex = coercomp(options, IO_CTL, CH_CTL))) {
    ;           CICLEWARN(1, "Not compatible CTL");
    ;   }   }
    ;   /**************************
    ;    *    IN *
    ;    **************************/
    ;   if  (a_cmd->flags & PGOBLIN_ISIN) {
    ;       if  (CH_IN & PGOBLIN_OUTSET) {
    ;           if  (r[PGO_CIN].r) {
                    if  ((ex = cleareg(options, IO_IN, PGOBLIN_PATHPARM | PGOBLIN_MIFEDES))) {
    ;                   CICLEWARN(2, "Not clear path|mife IN");
    ;           }   }
    ;       } else if (r[PGO_CIN].r) {
    ;           if  (  (r[PGO_CIN].r == r[PGO_COUT].r)
                    && (CH_IN & CH_OUT & (PGOBLIN_PATHPARM | PGOBLIN_TEXTPARM | PGOBLIN_PQRESULT))
                    ) {
    ;               ifBLIN_QV2(options->flags)
                        warnx( "cmd[%d]=[%03X(%s)]    \n"
                             , options->ppoint, cmd, a_cmd->str
                             );
    ;           }
    ;           if  (!(IO_IN->flags & PGOBLIN_ANYIVL)) {
    ;               options->exeflags |= PGOBLIN_FORKIN;
    ;           } else if ((ex = coercion(options, IO_IN, CH_IN))) {
    ;               CICLEWARN(2, "No coercion IN");
    ;           }
    ;       } else if ((ex = coercomp(options, IO_IN, CH_IN))) {
    ;           CICLEWARN(2, "Not compatible IN");
    ;   }   }
    ;   /***************************
    ;    *   CONN *
    ;    ***************************/
    ;   if  (a_cmd->flags & PGOBLIN_CONNECT) {
    ;       if  (CH_CON & PGOBLIN_OUTSET) {
    ;           PQfinish(r[PGO_CCON].g.c->conn);
    ;           r[PGO_CCON].g.c->conn = NULL;
    ;       } else if ((ex = pgoblin_shurecon(options, r[PGO_CCON].r))) {
    ;           CICLEWARN(3, "Not connected");
    ;   }   }
    ;   /**************************
    ;    *   JOB *
    ;    **************************/
    ;   if  (a_cmd->flags & PGOBLIN_JOB) {
    ;       if  (  (CH_JOB & PGOBLIN_OUTSET)
                && !(*r[PGO_CJOB].g.j)
                && !(*r[PGO_CJOB].g.j = mular_create(MULAR_ZERO, 3, 12, u12))
                ) {
    ;           ex = EX_SOFTWARE;
    ;           CICLEERR("No JOB");
    ;           continue;
    ;   }   }
    ;   /****************************
    ;    *   STYLE *
    ;    ****************************/
    ;   if  (a_cmd->flags & PGOBLIN_STYLE) {
    ;       /* When style come in */
    ;   }
    ;   /******************************
    ;    *     *
    ;    ******************************/
    ;   if  ((options->exeflags & PGOBLIN_FORKIN) && (ex = pipe(options->pip2))) {
    ;           CICLEERR("pipe #2");
    ;           close(options->pip2[0]);
    ;           close(options->pip2[1]);
    ;           continue;
    ;   }
    ;   if  ((options->exeflags & PGOBLIN_FORKCTL) && (ex = pipe(options->pip1))) {
    ;           CICLEERR("pipe #1");
    ;           close(options->pip1[0]);
    ;           close(options->pip1[1]);
    ;           continue;
    ;   }
    ;   if  ((options->exeflags & PGOBLIN_FORKOUT) && (ex = pipe(options->pip0))) {
    ;           CICLEERR("pipe #0");
    ;           close(options->pip0[0]);
    ;           close(options->pip0[1]);
    ;           continue;
    ;   }
    ;   ifBLIN_QV4(options->flags) {
            DUMPIOREGS;
            fprintf(stderr, "exeflags=%08X\n", options->exeflags);
    ;   }
    ;   /******************************
    ;    *            *
    ;    ******************************/
    ;   if  (  (cmd == PGOBLIN_fork) || (cmd == PGOBLIN_exec)
            || !(options->exeflags & (PGOBLIN_FORKIN | PGOBLIN_FORKCTL | PGOBLIN_FORKOUT))
            ) {
    ;   ;   ifBLIN_QV3(options->flags)
                fprintf(stderr, "e%06X DO c%03X %s\n", options->exeflags, cmd, a_cmd->str);
    ;       ex = (*(a_cmd->car))(options, r);
    ;   } else {
    ;       ifBLIN_QV3(options->flags)
                fprintf(stderr, "e%06X FORK c%03X %s\n", options->exeflags, cmd, a_cmd->str);
    ;   ;   if  ((ex = fork()) < 0) {
    ;   ;       CICLEERR("fork #1");
    ;   ;       continue;
    ;   ;   } else if (ex == 0) {
    ;   ;       /*********************************
    ;   ;        * ,    *
    ;   ;   ;    *********************************/
    ;   ;   ;   if  ((ex = pgoblin_connown(options))) exit(ex);
    ;   ;   ;   options->exeflags |= PGOBLIN_FORKED;
    ;   ;   ;   if  (options->exeflags & PGOBLIN_FORKIN) {
    ;   ;   ;       if  ((ex = pgoblin_inx(options, IO_IN, options->pip2, 1))) exit(ex);
    ;   ;   ;   }
    ;   ;   ;   if  (options->exeflags & PGOBLIN_FORKCTL) {
    ;   ;   ;       if  ((ex = pgoblin_inx(options, IO_CTL, options->pip1, 1))) exit(ex);
    ;   ;   ;   }
    ;   ;   ;   if  (options->exeflags & PGOBLIN_FORKOUT) {
    ;   ;   ;       if  ((ex = pgoblin_outx(options, IO_OUT, options->pip0, 1))) exit(ex);
    ;   ;       }
    ;   ;       exit((ex = (*(a_cmd->car))(options, r)));
    ;   ;   } else {
                pgoblin_job *j;

    ;   ;       j = mular_add(options->job[0]);
    ;   ;       j->cmdn = options->ppoint;
    ;   ;       j->pid = ex;
    ;   ;       j->status = PGOBLIN_NOSTATUS;
    ;   ;       ex = 0;
    ;   ;       if  (options->exeflags & PGOBLIN_FORKIN) {
    ;   ;           if  ((ex = pgoblin_outx(options, IO_IN, options->pip2, 1))) continue;
    ;   ;       }
    ;   ;       if  (options->exeflags & PGOBLIN_FORKCTL) {
    ;   ;           if  ((ex = pgoblin_outx(options, IO_CTL, options->pip1, 1))) continue;
    ;   ;       }
    ;   ;       if  (options->exeflags & PGOBLIN_FORKOUT) {
    ;   ;           IO_OUT->pid = j->pid;
    ;   ;           close(options->pip0[1]);
    ;   ;           if  (options->flags & PGOBLIN_OLDSTYLE && r[PGO_COUT].r == 1) {
    ;   ;               if  (IO_RG0.mife) mife_close(IO_RG0.mife);
    ;   ;               if  (!(IO_RG0.mife = mife_opef(MIFE_FULL, options->pip0[0]))) {
    ;   ;                   ex = EX_NOINPUT;
    ;   ;                   CICLEERR("Old style mife_opef");
    ;   ;               } else {
    ;   ;                   IO_RG0.offset = 0;
    ;   ;                   IO_RG0.flags |= PGOBLIN_MIFEDES | PGOBLIN_FREEMIFE;
    ;   ;               }
    ;   ;           } else {
    ;   ;               if  (!(IO_OUT->mife = mife_opef(MIFE_FULL, options->pip0[0]))) {
    ;   ;                   ex = EX_NOINPUT;
    ;   ;                   CICLEERR("New style mife_opef");
    ;                   } else {
    ;                       IO_OUT->offset = 0;
    ;                       IO_OUT->flags |= PGOBLIN_MIFEDES | PGOBLIN_FREEMIFE;
    ;   }   }   }   }   }
    ;   if  (ex) continue;
    ;   /*****************************
    ;    *  IO_OUT  *
    ;    *****************************/
    ;   if  (a_cmd->flags & PGOBLIN_OUTSET) {
    ;       ifBLIN_QV4(options->flags) {
    ;           fprintf(stderr, " %08X ", options->exeflags);
    ;   ;       dump_ioreg(options, r[PGO_COUT].r);
    ;   ;   }
    ;   ;   if  (options->exeflags & PGOBLIN_FREEPATH) {
                off_t off;
                mife_descriptor *md;
                char *cont;
                ssize_t bl;

    ;   ;       if  (!(IO_OUT->path)) {
    ;   ;           CICLEWARN(0, "unload no path");
    ;   ;       }
    ;   ;       md = mife_open(kukMIFE_BUFL(PGOBLIN_RAWIOBUFL) | MIFE_BUFX, IO_OUT->path);
    ;   ;       if  (!md) {
    ;   ;           ex = EX_IOERR;
    ;   ;           CICLEERR("unload mife open");
    ;   ;           continue;
    ;   ;       }
    ;   ;       for (off = 0; !(md->flag & MIFE_EOFL); off += bl) {
    ;   ;           if  ((bl = mife_read(md, 1 << (PGOBLIN_RAWIOBUFL + 16), off)) < 0) {
    ;   ;               ifBLIN_QV1(options->flags) warn("unload mife read %s:", IO_OUT->path);
    ;   ;               ex = EX_IOERR;
    ;   ;               break;
    ;   ;           }
    ;   ;           if  (!(cont = mife_get(md, off))) {
    ;   ;               ifBLIN_QV1(options->flags) warn("unload mife get %s:", IO_OUT->path);
    ;   ;               ex = EX_IOERR;
    ;   ;               break;
    ;   ;           }
    ;   ;           if  (IO_OUT->wri(IO_OUT->onu, cont, bl) < 0) {
    ;   ;               ifBLIN_QV1(options->flags) warn("unload mife write");
    ;   ;               ex = EX_IOERR;
    ;   ;               break;
    ;   ;       }   }
    ;   ;       if  (ex) continue;
    ;   ;       if  (mife_close(md)) {
    ;   ;           ex = EX_IOERR;
    ;   ;           CICLEERR("unload close");
    ;   ;           continue;
    ;   ;       }
    ;   ;       ex = cleareg(options, IO_OUT, PGOBLIN_PATHPARM);
    ;   ;       if  (ex) CICLEWARN(0, "IO_OUT->path not cleared");
    ;   ;   }
    ;   ;   if  (options->exeflags & PGOBLIN_FREETEXT) {
    ;   ;       if  (!(IO_OUT->text)) {
    ;   ;           CICLEWARN(0, "unload no text");
    ;   ;       }
    ;   ;       if  ((IO_OUT->flags & PGOBLIN_BINPARM) && !IO_OUT->length)
                    IO_OUT->length = strlen(IO_OUT->text);
    ;   ;       ex = IO_OUT->wri(IO_OUT->onu, IO_OUT->text, IO_OUT->length);
    ;   ;       if  (ex < 0) {
    ;   ;           ex = EX_IOERR;
    ;   ;           CICLEERR("unload text write");
    ;   ;           continue;
    ;   ;       }
    ;   ;       ex = cleareg(options, IO_OUT, PGOBLIN_TEXTPARM);
    ;   ;       if  (ex) CICLEWARN(0, "IO_OUT->text not cleared");
    ;   ;   }
    ;   ;   if  (options->exeflags & PGOBLIN_FREEPQRE) {
                int g, c;

    ;   ;       if  (!(IO_OUT->pq)) {
    ;   ;           CICLEWARN(0, "unload no pq");
    ;   ;       }
    ;   ;       for (g = 0; g < PQntuples(IO_OUT->pq) && ex >= 0; g++) {
    ;   ;           for (c = 0; c < PQnfields(IO_OUT->pq) && ex >= 0; c++) {
    ;   ;               if  (!PQgetisnull(IO_OUT->pq, g, c))
    ;                       ex = IO_OUT->wri( IO_OUT->onu
                                            , PQgetvalue(IO_OUT->pq, g, c)
                                            , PQgetlength(IO_OUT->pq, g, c)
                                            );
    ;   ;       }   }
    ;   ;       if  (ex < 0) {
    ;   ;           ex = EX_IOERR;
    ;   ;           CICLEERR("unload pq");
    ;   ;           continue;
    ;   ;       }
    ;   ;       ex = cleareg(options, IO_OUT, PGOBLIN_PQRESULT);
    ;   ;       if  (ex) CICLEWARN(0, "IO_OUT->pq not cleared");
    ;   ;   }
    ;   ;   if  (r[PGO_COUT].r == 0) {
    ;   ;       if  (IO_OUT->path && !(IO_OUT->flags & PGOBLIN_PATHPARM)) {
    ;   ;           ex = cleareg(options, IO_OUT, PGOBLIN_PATHPARM);
    ;   ;           if  (ex) CICLEWARN(0, "0 IO_OUT->path not cleared");
    ;   ;       }
    ;   ;       if  (IO_OUT->text && !(IO_OUT->flags & PGOBLIN_TEXTPARM)) {
    ;   ;           ex = cleareg(options, IO_OUT, PGOBLIN_TEXTPARM);
    ;   ;           if  (ex) CICLEWARN(0, "0 IO_OUT->text not cleared");
    ;   ;       }
    ;   ;       if  (IO_OUT->pq && !(IO_OUT->flags & PGOBLIN_PQRESULT)) {
    ;   ;           ex = cleareg(options, IO_OUT, PGOBLIN_PQRESULT);
    ;   ;           if  (ex) CICLEWARN(0, "0 IO_OUT->pq not cleared");
    ;   ;       }
    ;   ;       if  (IO_OUT->mife && !(IO_OUT->flags & PGOBLIN_MIFEDES)) {
    ;   ;           ifBLIN_QV1(options->flags) warnx("Clear mife in io[0]");
    ;   ;           ex = cleareg(options, IO_OUT, PGOBLIN_MIFEDES);
    ;   ;           if  (ex) CICLEWARN(0, "0 IO_OUT->mife not cleared");
    ;   ;       }
    ;   ;       if  (IO_OUT->wri && !(IO_OUT->flags & PGOBLIN_OUTSET)) {
    ;   ;           ifBLIN_QV1(options->flags) warnx("Clear wri in io[0]");
    ;   ;           ex = cleareg(options, IO_OUT, PGOBLIN_OUTSET);
    ;   ;           if  (ex) CICLEWARN(0, "0 IO_OUT->wri not cleared");
    ;   ;   }   }
    ;   ;   ifBLIN_QV4(options->flags) {
    ;           fprintf(stderr, "  ");
    ;           dump_ioreg(options, r[PGO_COUT].r);
    ;   }   }
    ;   if  ((a_cmd->flags & PGOBLIN_CTL) && !(r[PGO_CCTL].r) && (CH_CTL & PGOBLIN_TEXTPARM)) {
    ;       IO_CTL->text = NULL;
        }
        ifBLIN_QV1(options->flags) iorassert(options, cmd);
    }
    if  (!ex && (options->exeflags & PGOBLIN_ERROR)) ex = -1;
    ifBLIN_QV3(options->flags) fprintf(stderr, "-pgoblin_execute %d\n", ex);
    return(ex);
};
